#!/usr/bin/python3

#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#
"""The Python implementation of the gRPC doca_dpi_server.DocaDPI client."""

import logging
import click
import time
import sys
from scapy.all import *
import grpc
import atexit

import common_pb2 as gen_common
import doca_dpi_types_pb2 as dpi_types
import doca_dpi_pb2 as gen_pbuf
import doca_dpi_pb2_grpc as gen_grpc

LIB_NAME = 'DOCA DPI'
FULL_LIB_NAME = LIB_NAME + ' gRPC Client'
RTE_PKTMBUF_HEADROOM = 128

logger = logging.getLogger(FULL_LIB_NAME)
logger.setLevel(logging.INFO)

def handle_error_info(err, func_name):
	if err.error_code < 0:
		if err.HasField("err_msg"):
			err_msg = "code: %d: " % err.error_code
			raise Exception(err_msg + err.err_msg)
		else:
			raise Exception("Unknown error at doca_dpi, code: %d, at %s" % (err.error_code, func_name))


class DocaDpiClient(object):
	"""
	Client for gRPC doca DPI API
	"""
	def __init__(self, server_address):
		logger.info(f'Connecting to the {LIB_NAME} gRPC server on the DPU: {server_address}')
		# Instantiate a channel
		self.channel = grpc.insecure_channel(server_address)
		# Bind the client and the server
		try:
			self.stub = gen_grpc.DocaDpiServiceStub(self.channel)
		except RecursionError as e:
			logger.error('Failed to connect to the gRPC server on the DPU')
		# Destroy serve dpi context on exit
		atexit.register(self._doca_dpi_destroy)

	def doca_dpi_load_signatures(self, path_to_cdo, is_on_dpu=True):
		logger.debug(f'{LIB_NAME} Load signatures request')
		if is_on_dpu:
			request = gen_pbuf.DocaDpiLoadSignaturesParams(cdo_filename=path_to_cdo)
		else:
			with open(path_to_cdo, "rb") as f:
				request = gen_pbuf.DocaDpiLoadSignaturesParams(cdo_data=f.read())
		res = self.stub.DocaDpiLoadSignatures(request)
		handle_error_info(res, "doca_dpi_load_signatures")

	def doca_dpi_flow_create(self, l4_protocol, is_ipv4, src_ip, src_port, dst_ip, dst_port, ethertype):
		logger.debug(f'{LIB_NAME} Flow create request')
		if is_ipv4:
			dst_ip = reduce(lambda x,y: (x << 8) | y, map(int, dst_ip.split('.')))
			src_ip = reduce(lambda x,y: (x << 8) | y, map(int, src_ip.split('.')))
			parse_info = dpi_types.DocaDpiParsingInfo(ethertype=ethertype, l4_protocol=l4_protocol,
				l4_dst_port=dst_port, l4_src_port=src_port, dst_ipv4=dst_ip, src_ipv4=src_ip)
		else:
			src_ipv6 = dpi_types.In6Addr(ipv6_addr=src_ip)
			dst_ipv6 = dpi_types.In6Addr(ipv6_addr=dst_ip)
			parse_info = dpi_types.DocaDpiParsingInfo(ethertype=ethertype, l4_protocol=l4_protocol,
				l4_dst_port=dst_port, l4_src_port=src_port, dst_ipv6=dst_ipv6, src_ipv6=src_ipv6)
		request = gen_pbuf.DocaDpiFlowCreateParams(parsing_info=parse_info, uint16_dpi_q=0)
		res = self.stub.DocaDpiFlowCreate(request)
		handle_error_info(res.erreno, "doca_dpi_flow_create")
		return res.flow_context, res.result

	def doca_dpi_flow_destroy(self, ctx:dpi_types.DocaDpiFlowCtx):
		logger.debug(f'{LIB_NAME} Destroy flow context request')
		self.stub.DocaDpiFlowDestroy(gen_pbuf.DocaDpiFlowDestroyParams(context=ctx))

	def doca_dpi_flow_match_get(self, ctx:dpi_types.DocaDpiFlowCtx):
		logger.debug(f'{LIB_NAME} Flow match request')
		res = self.stub.DocaDpiFlowMatchGet(gen_pbuf.DocaDpiFlowMatchGetParams(context=ctx))
		handle_error_info(res.erreno, "doca_dpi_flow_match_get")
		return res.result

	def doca_dpi_enqueue(self, flow_ctx:dpi_types.DocaDpiFlowCtx, packets:dpi_types.DocaDpiGenericPacket,
			initiator, payload_offset, user_data=None):
		logger.debug(f'{LIB_NAME} Enqueue request')
		request = gen_pbuf.DocaDpiEnqueueParams(context=flow_ctx, packet=packets, initiator=initiator,
			payload_offset=payload_offset)
		if user_data is not None:
			request.user_data = user_data
		res = self.stub.DocaDpiEnqueue(request)
		handle_error_info(res, "doca_dpi_enqueue")
		return res.error_code

	def doca_dpi_dequeue(self, dpi_queue):
		logger.debug(f'{LIB_NAME} Dequeue request')
		request = gen_pbuf.DocaDpiDequeueParams(uint16_dpi_q=dpi_queue)
		res = self.stub.DocaDpiDequeue(request)
		handle_error_info(res.erreno, "doca_dpi_dequeue")
		return res.erreno.error_code, res.result

	def doca_dpi_signature_get(self, sign_id):
		logger.debug(f'{LIB_NAME} SignatureGet request')
		request = gen_pbuf.DocaDpiSignatureGetParams(sign_id=sign_id)
		res = self.stub.DocaDpiSignatureGet(request)
		handle_error_info(res.erreno, "doca_dpi_signature_get")
		return res.sign_data

	def doca_dpi_signatures_get(self):
		logger.debug(f'{LIB_NAME} SignaturesGet request')
		res = self.stub.DocaDpiSignaturesGet(gen_pbuf.DocaDpiSignaturesGetParams())
		handle_error_info(res.erreno, "doca_dpi_signatures_get")
		return res.signs

	def doca_dpi_stat_get(self, clear:bool):
		logger.debug(f'{LIB_NAME} StatGet request')
		request = gen_pbuf.DocaDpiStatGetParams(clear=clear)
		return self.stub.DocaDpiStatGet(request)

	def _doca_dpi_destroy(self):
		logger.debug(f'{LIB_NAME} Destroy request at destructor')
		self.stub.DocaDpiDestroy(gen_pbuf.DocaDpiDestroyParams())

class PacketToFlowContext(object):
	def __init__(self):
		self.map = {}

	def get_flow_ctx(self, packet:scapy.packet.Packet):
		l4proto = UDP if TCP not in packet else TCP
		key = (packet.proto, packet[IP].src, packet[l4proto].sport, packet[IP].dst, packet[l4proto].dport)
		return self.map[key] if key in self.map else None

	def add_flow_ctx(self, packet:scapy.packet.Packet, flow_ctx):
		l4proto = UDP if TCP not in packet else TCP
		dir1_key = (packet.proto, packet[IP].src, packet[l4proto].sport, packet[IP].dst, packet[l4proto].dport)
		self.map[dir1_key] = flow_ctx
		dir2_key = (packet.proto, packet[IP].dst, packet[l4proto].dport, packet[IP].src, packet[l4proto].sport)
		self.map[dir2_key] = flow_ctx

	def get_unique_values(self):
		unique_vals = {}
		for ctx in self.map.values():
			if ctx.unique_id not in unique_vals:
				unique_vals[ctx.unique_id] = ctx
		return unique_vals.values()


def create_flow(client_obj, packet:scapy.packet.Packet):
	l4proto = UDP if TCP not in packet else TCP
	packet.getlayer(scapy.layers.inet.IP).src  # ip src
	res = client_obj.doca_dpi_flow_create(packet.proto, packet[IP].version == 4, packet[IP].src,
		packet[l4proto].sport, packet[IP].dst, packet[l4proto].dport, packet.type)
	return res

def generate_DocaDpiGenericPacket(packet:scapy.packet.Packet):
	segment = bytes(packet)
	payload_offset = len(segment) - len(bytes(packet.lastlayer()))
	ret = dpi_types.DocaDpiGenericPacket(segment=segment)
	return ret, payload_offset

@click.group(help='DOCA DPI gRPC Client', invoke_without_command=True)
@click.argument('server_address')
@click.argument('signs_file_path')
@click.argument('is_signs_file_on_dpu', type=bool)
@click.option('-d', '--debug', is_flag=True, default=False)
def start_client(server_address, signs_file_path, is_signs_file_on_dpu, debug):
	# gRPC-generic commands and options
	if debug:
		logger.setLevel(logging.DEBUG)

	# Create fake packet
	dip="192.168.100.2"
	dmac="0c:42:a1:98:c7:a5"
	payload="you" * 668 + "tube" * 512
	packet=Ether(dst=dmac)/IP(dst=dip,id=12345)/UDP(sport=1500,dport=1501)/payload
	packet.show()
	payload="A" * 64
	packet2=Ether(dst=dmac)/IP(dst=dip,id=12345)/UDP(sport=1500,dport=1501)/payload
	packet2.show()
	packets = [packet2, packet, packet2]

	# Connection to the server
	grpc_server_port = gen_common.eNetworkPort.k_DocaDPIService
	if ':' not in server_address:
		server_ip = server_address
		server_address = f'{server_address}:{grpc_server_port}'
	else:
		server_ip = server_address[:server_address.find(':')]
	client = DocaDpiClient(server_address)

	### Using gRPC DPI ###
	# Loading signatures DB
	client.doca_dpi_load_signatures(signs_file_path, is_signs_file_on_dpu)
	signs = client.doca_dpi_signatures_get()
	con_track = PacketToFlowContext()

	# Printing signatures DB
	for sign_data in signs:
		logger.info("{ sig_id: %d name: %s }" % (sign_data.sig_id, sign_data.name))

	# Enqueue all packets
	for packet in packets:
		# We're only interested in packets with at least a L3 header
		if not packet.haslayer(scapy.layers.inet.IP):
			continue
		# Get flow id
		flow_ctx = con_track.get_flow_ctx(packet)
		if flow_ctx is None:
			flow_ctx, result = create_flow(client, packet)
			con_track.add_flow_ctx(packet, flow_ctx)
		# Check packet direction
		initiator = packet[IP].dst == server_ip
		# Parse packet to gRPC type
		grpc_pkt, payload_offset = generate_DocaDpiGenericPacket(packet)
		# Enqueue with nonsense user_data bytes
		status = client.doca_dpi_enqueue(flow_ctx, grpc_pkt, initiator, payload_offset,
			b"Test user_data\x56\x00\x03\x14\x15\x00\x00\x12")
		if status != 0:
			logger.error("Failed to enqueue a packet, error code: ", status)
			exit(-1)

	# Inspect Result
	for packet in packets:
		flow_ctx = con_track.get_flow_ctx(packet)
		status, result = client.doca_dpi_dequeue(0)
		if result.matched:
			match_data = client.doca_dpi_signature_get(result.info.sig_id)
			logger.info("Match for a flow was found on dequeue.")
			logger.info("	Signature  - " + match_data.name)
			logger.info("\tData of dequeued packet:")
			logger.info(result.pkt.segment)
			if result.HasField("user_data"):
				logger.info("\tUser data on matched packet:")
				logger.info(result.user_data)

	# Print statistics
	stats = client.doca_dpi_stat_get(False)
	logger.info("\ndpi stats {\n" + str(stats) + "}")

	# Testing flow match get
	for flow_ctx in con_track.get_unique_values():
		client.doca_dpi_flow_match_get(flow_ctx)
		if result.matched:
			match_data = client.doca_dpi_signature_get(result.info.sig_id)
			logger.info("Match for a flow was found on flow_match_get.")
			logger.info("	Signature  - " + match_data.name)

	# Destroying all flows
	for flow_ctx in con_track.get_unique_values():
		client.doca_dpi_flow_destroy(flow_ctx)

	logger.info('Finished Successfully')


if __name__ == '__main__':
	logging.basicConfig()
	start_client()
