#!/usr/bin/python3

#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
#
# This software product is a proprietary product of NVIDIA CORPORATION &
# AFFILIATES (the "Company") and all right, title, and interest in and to the
# software product, including all associated intellectual property rights, are
# and shall remain exclusively with the Company.
#
# This software product is governed by the End User License Agreement
# provided with the software product.
#
"""The Python implementation of the gRPC doca_flow_server.DocaFlow client."""

import logging
import click
import time
import sys

import grpc

import common_pb2 as gen_common
import doca_flow_service_pb2 as gen_pbuf
import doca_flow_service_pb2_grpc as gen_grpc

LIB_NAME = 'DOCA Flow'
FULL_LIB_NAME = LIB_NAME + ' gRPC Client'

logger = logging.getLogger(FULL_LIB_NAME)
logger.setLevel(logging.INFO)

class DocaFlowClient(object):
    """
    Client for gRPC doca flow API
    """
    def __init__(self, server_address):
        logger.info(f'Connecting to the {LIB_NAME} gRPC server on the DPU: {server_address}')
        # instantiate a channel
        self.channel = grpc.insecure_channel(server_address)
        # bind the client and the server
        try:
            self.stub = gen_grpc.DocaFlowStub(self.channel)
        except RecursionError as e:
                logger.error('Failed to connect to the gRPC server on the DPU')

    def doca_flow_init(self, total_sessions, queues, is_hairpin, aging):
        logger.debug(f'{LIB_NAME} Initialization request')
        request = gen_pbuf.DocaFlowCfg(total_sessions=total_sessions, queues=queues, is_hairpin=is_hairpin, aging=aging)
        return self.stub.DocaFlowInit(request)

    def doca_flow_port_start(self, port_id, type, devargs, priv_data_size):
        logger.debug(f'{LIB_NAME} Port start request')
        request = gen_pbuf.DocaFlowPortCfg(port_id=port_id, type=type, devargs=devargs, priv_data_size=priv_data_size)
        return self.stub.DocaFlowPortStart(request)

    def doca_flow_create_pipe(self, cfg, fwd):
        logger.debug(f'{LIB_NAME} Create pipe request')
        request = gen_pbuf.DocaFlowCreatePipeRequest(cfg=cfg, fwd=fwd)
        return self.stub.DocaFlowCreatePipe(request)

    def doca_flow_pipe_add_entry(self, port_id, pipe_queue, pipe_id, match, actions, monitor, fwd):
        logger.debug(f'{LIB_NAME} Add Entry to a pipe request')

        request = gen_pbuf.DocaFlowPipeAddEntryRequest(port_id=port_id, pipe_queue=pipe_queue, pipe_id=pipe_id,
                         match=match, actions=actions, monitor=monitor, fwd=fwd)
        start_time = time.time()
        add_entry_res =  self.stub.DocaFlowPipeAddEntry(request)
        end_time = time.time()
        curr_rtt = end_time - start_time
        curr_rate = 1/curr_rtt
        return (add_entry_res,curr_rate)

    def doca_flow_pipe_rm_entry(self, pipe_queue, entry_id, pipe_id, port_id):
        logger.debug(f'{LIB_NAME} Remove Entry of a pipe request')
        request = gen_pbuf.DocaFlowRemoveEntryRequest(pipe_queue=pipe_queue, entry_id=entry_id,
                         pipe_id=pipe_id, port_id=port_id)
        return self.stub.DocaFlowRemoveEntry(request)

    def doca_flow_flush_pipes(self, port_id):
        logger.debug(f'{LIB_NAME} flushing pipes of port {port_id} request')
        request = gen_pbuf.DocaFlowPort(port_id=port_id)
        return self.stub.DocaFlowFlushPipes(request)

    def doca_flow_dump_pipe(self, port_id):
        logger.debug(f'{LIB_NAME} dump pipes of port {port_id} request')
        request = gen_pbuf.DocaFlowPort(port_id=port_id)
        return self.stub.DocaFlowDumpPipe(request)

    def doca_flow_destroy_pipe(self, port_id, pipe_id):
        logger.debug(f'{LIB_NAME} destroying pipe request')
        request = gen_pbuf.DocaFlowDestroyPipeRequest(port_id=port_id,pipe_id=pipe_id)
        return self.stub.DocaFlowDestroyPipe(request)

    def doca_flow_destroy_port(self, port_id):
        logger.debug(f'{LIB_NAME} destroy pipes of port {port_id} request')
        request = gen_pbuf.DocaFlowPort(port_id=port_id)
        return self.stub.DocaFlowDestroyPort(request)

    def doca_flow_destroy(self):
        logger.debug(f'{LIB_NAME} destroy request')
        request = gen_pbuf.DocaFlowDestroyRequest()
        return self.stub.DocaFlowDestroyPort(request)

@click.group(help='DOCA Flow gRPC Client', invoke_without_command=True)
@click.argument('server_address')
@click.argument('number_of_insertions')
@click.option('-d', '--debug', is_flag=True, default=False)
def start_client(server_address, number_of_insertions, debug):
    # This function starts a very simple DOCA FLOW gRPC client.
    # The client starts the DOCA FLOW API. It initializes two ports and create one pipe with one entry for each port.
    # Each pipe/entry contains the same components.
    # That is they matches packets of ip4 type and destination port 53. Upon match it changes the destination
    # port from 53 to 44 (this is the action component of the pipe/entry).
    # In addition, the matched packets are forwarded to the ARM (the FORWARDING component).

    # gRPC-generic commands and options
    if debug:
        logger.setLevel(logging.DEBUG)

    grpc_server_port = gen_common.eNetworkPort.k_DocaFlowService
    if ':' not in server_address:
        server_address = f'{server_address}:{grpc_server_port}'


    client = DocaFlowClient(server_address)
    doca_flow_init_response = client.doca_flow_init(total_sessions=2, queues=8, is_hairpin=False, aging=False)
    if not doca_flow_init_response.success:
        logger.debug(doca_flow_init_response.error.err_message)

    ports_nb = 2
    ports = [ i for i in range(0, ports_nb)]
    insertions = int(number_of_insertions)
    entries_idx_arr = [ i for i in range(0,insertions)]
    # insertion rate average for each port.
    irate_avg_arr = [ 0 for i in range(0, ports_nb)]

    for port_id in ports:
        # starting DOCA Flow ports.
        ports_start_response = client.doca_flow_port_start(port_id=port_id, type=gen_pbuf.DocaFlowPortType.PORT_DPDK_BY_ID, devargs=str(port_id), priv_data_size=0)

        # creating a pipe
        name = "simple client pipe 1"
        ip_type = gen_pbuf.DocaFlowIPtype.IP4_ADDR
        udp_ip_address = gen_pbuf.DocaFlowIPAddress(type=ip_type)

        # match for UDP packets with changeable out_dst_port
        match = gen_pbuf.DocaFlowMatch(out_dst_ip=udp_ip_address, out_l4_type=17,out_dst_port=0xffff )

        actions = gen_pbuf.DocaFlowActions(mod_dst_port=44)

        # pipe configuarion initialization
        cfg = gen_pbuf.DocaFlowPipeCfg(is_root=True,name=name, port_id=port_id, match=match,
                 match_mask=None , actions=actions, monitor=None)

        # pipe's forward component initialization
        rss_queues= [i for i in range(0,7)]
        # bit manipulation , should be equal ETH_RSS_UDP
        rss_flags = 41868 # RSS flags
        fwd = gen_pbuf.DocaFlowFwd(type=gen_pbuf.DocaFlowFwdType.FWD_RSS, rss_flags=rss_flags, rss_queues=rss_queues,
                          num_of_queues=8, rss_mark=0)

        # pipe creation
        pipe_response = client.doca_flow_create_pipe(cfg, fwd)
        min_rate = 0
        for entry_idx in entries_idx_arr:

            # match for DNS packets
            match = gen_pbuf.DocaFlowMatch(out_dst_port=53+entry_idx)

            # adding entries to the pipe
            res = client.doca_flow_pipe_add_entry(port_id, 0, pipe_response.pipe_id,
                   match, actions, None, None)

            add_entry_response =res[0]
            curr_rate = res[1]
            # performance calculations
            min_rate = min (curr_rate, sys.maxsize)
            irate_avg_arr[port_id] += curr_rate

            if add_entry_response.success:
                logger.debug(f'entry idx: {entry_idx} added to the pipe for port {port_id} successfully with insertion rate {curr_rate}')
            else:
                logger.debug(add_entry_response.error.err_message)

        irate_avg_arr[port_id] /= insertions
        logger.info(f'The insertion rate average for port {port_id} is {irate_avg_arr[port_id]}')

    irate_avg = sum(irate_avg_arr) / ports_nb
    logger.info(f'Insertion rate AVERAGE is {irate_avg} per second ')
    logger.info(f'MINIMUM insertion rate is {min_rate} per second ')

    for port_id in ports:
        response = client.doca_flow_dump_pipe(port_id)
        logger.info(response.dump_pipes)

    client.doca_flow_destroy()
    logger.info('Finished Successfully')

if __name__ == '__main__':

    logging.basicConfig()
    start_client()
